from . import websocket
from .tests.defaults import MIN_PYTHON_MINOR, MIN_PYTHON_MAJOR
from .history import TD_hist
from .analytics import TD_analytics
from .websocket import TD_live

__version__ = '6.0.0-beta.3'
